#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
import pyqtgraph.opengl as gl
from PySide6.QtGui import QQuaternion
from PySide6.QtWidgets import QDialog

from pyminflux.ui.ui_plotter_3d import Ui_Plotter3D


class Plotter3D(QDialog, Ui_Plotter3D):
    def __init__(self, parent=None):
        """Constructor."""

        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_Plotter3D()
        self.ui.setupUi(self)

        # Initialize the OpenGL View
        self.view = gl.GLViewWidget()
        self.ui.mainLayout.addWidget(self.view)

        # Keep a reference to the grid and the plot
        self.grid = None
        self.scatter = None

        # Set color
        self.color = (1.0, 1.0, 1.0, 0.5)

        # Only hide on close by default
        self.hide_on_close = True

    def plot(self, coords):
        """Plot the data in 3D."""

        if coords.shape[1] != 3:
            raise ValueError("coords must be an (Nx3) array.")

        # Calculate ranges and positions
        min_x, max_x = coords[:, 0].min(), coords[:, 0].max()
        min_y, max_y = coords[:, 1].min(), coords[:, 1].max()
        min_z, max_z = coords[:, 2].min(), coords[:, 2].max()
        range_x = max_x - min_x
        range_y = max_y - min_y
        range_z = 0.0
        spacing_x = range_x / 25
        spacing_y = range_y / 25
        spacing_z = 0.0
        center_x = 0.5 * (min_x + max_x)
        center_y = 0.5 * (min_y + max_y)
        center_z = 0.0
        distance = np.max((range_x, range_y, range_z))

        # Remove items if they already exist
        if self.grid is not None:
            self.view.removeItem(self.grid)
            self.grid.deleteLater()

        if self.scatter is not None:
            self.view.removeItem(self.scatter)
            self.scatter.deleteLater()

        # Create a grid and center it at the center of mass of the point cloud
        self.grid = gl.GLGridItem()
        self.grid.setSize(range_x, range_y, range_z)
        self.grid.setSpacing(spacing_x, spacing_y, spacing_z)
        self.view.addItem(self.grid)

        # Add the points
        self.scatter = gl.GLScatterPlotItem(
            pos=coords, size=5.0, color=self.color, pxMode=True
        )
        self.scatter.translate(-center_x, -center_y, -center_z)
        self.view.addItem(self.scatter)

        # Set the camera
        self.view.opts["distance"] = distance
        self.view.opts["fov"] = 60
        self.view.opts["elevation"] = 90
        self.view.opts["azimuth"] = 270
        self.view.opts["rotation"] = QQuaternion(1.000000, 0.000000, 0.000000, 0.000000)

        # Show
        self.view.show()

    def closeEvent(self, ev):
        """Hide the dialog instead of closing it."""
        if self.hide_on_close:
            ev.ignore()
            self.hide()
        else:
            ev.accept()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np

from pyminflux.analysis import ideal_hist_bins


def calculate_common_bins(first_array_values, second_array_values):
    """Calculate bins commons to both sets of values.

    Parameters
    ----------

    first_array_values: np.ndarray
        First array of values. It may contain NaNs.

    second_array_values: np.ndarray
        Second array of values. It may contain NaNs.

    Returns
    -------

    bin_edges: np.ndarray
        Array of bin edges (compatible with NumPy.histogram())

    bin_centers: np.ndarray
        Array of bin centers

    bin_width: float
        Bin width.
    """
    num_first = np.sum(np.logical_not(np.isnan(first_array_values)))
    num_second = np.sum(np.logical_not(np.isnan(second_array_values)))
    if num_first == 0 and num_second == 0:
        raise ValueError("No usable data.")
    elif num_first >= num_second:
        bin_edges, bin_centers, bin_width = ideal_hist_bins(first_array_values)
    elif num_first < num_second:
        bin_edges, bin_centers, bin_width = ideal_hist_bins(second_array_values)
    else:
        raise ValueError("Unexpected case.")

    return bin_edges, bin_centers, bin_width


def print_summary_statistics(values: np.ndarray):
    """Prints summary statistics.

    Parameters
    ----------

    values: np.ndarray
        Array of values. It may contain NaNs.
    """

    # Count number of valid entries
    num_valid = np.logical_not(np.isnan(values)).sum()

    if num_valid == 0:
        print("Nothing to summarize.")
        return

    print(
        f"Num valid entries = {num_valid}/{len(values)} ({100 * num_valid / len(values):.2f}%)\n"
        f"Min               = {np.nanmin(values)}\n"
        f"25% percentile    = {np.nanpercentile(values, 25)}\n"
        f"Median            = {np.nanmedian(values)}\n"
        f"Mean              = {np.nanmean(values):.2f}\n"
        f"75% percentile    = {np.nanpercentile(values, 75)}\n"
        f"Max               = {np.nanmax(values)}"
    )



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#


class Singleton(type):
    """Metaclass to implement the singleton pattern for classes."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-

#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

################################################################################
## Form generated from reading UI file 'plotter_3d.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import QApplication, QDialog, QGridLayout, QSizePolicy, QWidget


class Ui_Plotter3D(object):
    def setupUi(self, Plotter3D):
        if not Plotter3D.objectName():
            Plotter3D.setObjectName("Plotter3D")
        Plotter3D.resize(800, 600)
        self.gridLayout_2 = QGridLayout(Plotter3D)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.mainLayout = QGridLayout()
        self.mainLayout.setObjectName("mainLayout")

        self.gridLayout_2.addLayout(self.mainLayout, 0, 0, 1, 1)

        self.retranslateUi(Plotter3D)

        QMetaObject.connectSlotsByName(Plotter3D)

    # setupUi

    def retranslateUi(self, Plotter3D):
        Plotter3D.setWindowTitle(
            QCoreApplication.translate("Plotter3D", "3D plotter", None)
        )

    # retranslateUi
